@@ -106,6 +106,28 @@ struct msc_a *msc_a_fi_priv(struct osmo_fsm_inst *fi)
 	return fi->priv;
 }
 
+bool msc_a_is_ciphering_to_be_attempted(const struct msc_a *msc_a)
+{
+	struct gsm_network *net = msc_a_net(msc_a);
+	bool is_utran = (msc_a->c.ran->type == OSMO_RAT_UTRAN_IU);
+	if (is_utran)
+		return net->uea_encryption_mask > (1 << OSMO_UTRAN_UEA0);
+	else
+		return net->a5_encryption_mask > 0x1;
+}
+
+bool msc_a_is_ciphering_required(const struct msc_a *msc_a)
+{
+	struct gsm_network *net = msc_a_net(msc_a);
+	bool is_utran = (msc_a->c.ran->type == OSMO_RAT_UTRAN_IU);
+	if (is_utran)
+		return net->uea_encryption_mask
+			&& ((net->uea_encryption_mask & (1 << OSMO_UTRAN_UEA0)) == 0);
+	else
+		return net->a5_encryption_mask
+			&& ((net->a5_encryption_mask & 0x1) == 0);
+}
+
 static void update_counters(struct osmo_fsm_inst *fi, bool conn_accepted)
 {
 	struct msc_a *msc_a = fi->priv;
@@ -1301,13 +1323,15 @@ static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct
 {
 	struct gsm_trans *cc_trans = msc_a->cc.active_trans;
 	struct rtp_stream *rtps_to_ran = msc_a->cc.call_leg ? msc_a->cc.call_leg->rtp[RTP_TO_RAN] : NULL;
+	const enum mgcp_codecs *codec_if_known = ac->assignment_complete.codec_present ?
+							&ac->assignment_complete.codec : NULL;
 
 	if (!rtps_to_ran) {
 		LOG_MSC_A(msc_a, LOGL_ERROR, "Rx Assignment Complete, but no RTP stream is set up\n");
 		return;
 	}
 	if (!cc_trans) {
-		LOG_MSC_A(msc_a, LOGL_ERROR, "Rx Assignment Complete, but CC transaction is active\n");
+		LOG_MSC_A(msc_a, LOGL_ERROR, "Rx Assignment Complete, but no CC transaction is active\n");
 		return;
 	}
 
@@ -1319,7 +1343,8 @@ static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct
 	}
 
 	/* Update RAN-side endpoint CI: */
-	rtp_stream_set_codec(rtps_to_ran, ac->assignment_complete.codec);
+	if (codec_if_known)
+		rtp_stream_set_codec(rtps_to_ran, *codec_if_known);
 	rtp_stream_set_remote_addr(rtps_to_ran, &ac->assignment_complete.remote_rtp);
 	if (rtps_to_ran->use_osmux)
 		rtp_stream_set_remote_osmux_cid(rtps_to_ran,
@@ -1334,7 +1359,7 @@ static void msc_a_up_call_assignment_complete(struct msc_a *msc_a, const struct
 	 * - the Assignment has chosen a speech codec
 	 * go on to create the CN side RTP stream's CI. */
 	if (call_leg_ensure_ci(msc_a->cc.call_leg, RTP_TO_CN, cc_trans->callref, cc_trans,
-			       &ac->assignment_complete.codec, NULL)) {
+			       codec_if_known, NULL)) {
 		LOG_MSC_A_CAT(msc_a, DCC, LOGL_ERROR, "Error creating MGW CI towards CN\n");
 		call_leg_release(msc_a->cc.call_leg);
 		return;
@@ -1649,12 +1674,9 @@ int _msc_a_msg_down(struct msc_a *msc_a, enum msc_role to_role, uint32_t to_role
 		.an_proto = msc_a->c.ran->an_proto,
 		.msg = msc_role_ran_encode(msc_a->c.fi, ran_msg),
 	};
-	int rc;
 	if (!an_apdu.msg)
 		return -EIO;
-	rc = _msub_role_dispatch(msc_a->c.msub, to_role, to_role_event, &an_apdu, file, line);
-	msgb_free(an_apdu.msg);
-	return rc;
+	return _msub_role_dispatch(msc_a->c.msub, to_role, to_role_event, &an_apdu, file, line);
 }
 
 int msc_a_tx_dtap_to_i(struct msc_a *msc_a, struct msgb *dtap)
@@ -1718,6 +1740,7 @@ static int msc_a_start_assignment(struct msc_a *msc_a, struct gsm_trans *cc_tran
 	struct call_leg *cl = msc_a->cc.call_leg;
 	struct msc_i *msc_i = msc_a_msc_i(msc_a);
 	struct gsm_network *net = msc_a_net(msc_a);
+	enum mgcp_codecs codec, *codec_ptr;
 
 	OSMO_ASSERT(!msc_a->cc.active_trans);
 	msc_a->cc.active_trans = cc_trans;
@@ -1752,8 +1775,14 @@ static int msc_a_start_assignment(struct msc_a *msc_a, struct gsm_trans *cc_tran
 	 * If the local address is already known, then immediately trigger. */
 	if (call_leg_local_ip(cl, RTP_TO_RAN))
 		return osmo_fsm_inst_dispatch(msc_a->c.fi, MSC_EV_CALL_LEG_RTP_LOCAL_ADDR_AVAILABLE, cl->rtp[RTP_TO_RAN]);
-	else
-		return call_leg_ensure_ci(msc_a->cc.call_leg, RTP_TO_RAN, cc_trans->callref, cc_trans, NULL, NULL);
+
+	if (msc_a->c.ran->type == OSMO_RAT_UTRAN_IU) {
+		codec = CODEC_IUFP;
+		codec_ptr = &codec;
+	} else {
+		codec_ptr = NULL;
+	}
+	return call_leg_ensure_ci(msc_a->cc.call_leg, RTP_TO_RAN, cc_trans->callref, cc_trans, codec_ptr, NULL);
 }
 
 int msc_a_try_call_assignment(struct gsm_trans *cc_trans)
@@ -1775,6 +1804,10 @@ int msc_a_try_call_assignment(struct gsm_trans *cc_trans)
 	return msc_a_start_assignment(msc_a, cc_trans);
 }
 
+/* Map CM Service type to use token.
+ * Given a CM Service type, return a matching token intended for osmo_use_count.
+ * For unknown service type, return NULL.
+ */
 const char *msc_a_cm_service_type_to_use(enum osmo_cm_service_type cm_service_type)
 {
 	switch (cm_service_type) {
