@@ -30,31 +30,89 @@
 #include <osmocom/msc/debug.h>
 #include <osmocom/msc/sdp_msg.h>
 
+bool sdp_audio_codec_is_set(const struct sdp_audio_codec *a)
+{
+	return a && a->subtype_name[0];
+}
+
 /* Compare name, rate and fmtp, returning typical cmp result: 0 on match, and -1 / 1 on mismatch.
- * Do *not* compare the payload_type number.
+ * If cmp_fmtp is false, do *not* compare the fmtp string; if true, compare fmtp 1:1 as strings.
+ * If cmp_payload_type is false, do *not* compare the payload_type number.
  * The fmtp is only string-compared -- e.g. if AMR parameters appear in a different order, it amounts to a mismatch even
  * though all parameters are the same. */
-int sdp_audio_codec_cmp(const struct sdp_audio_codec *a, const struct sdp_audio_codec *b)
+int sdp_audio_codec_cmp(const struct sdp_audio_codec *a, const struct sdp_audio_codec *b,
+			bool cmp_fmtp, bool cmp_payload_type)
 {
-	int rc;
+	int cmp;
 	if (a == b)
 		return 0;
 	if (!a)
 		return -1;
 	if (!b)
 		return 1;
-	rc = strncmp(a->subtype_name, b->subtype_name, sizeof(a->subtype_name));
-	if (rc)
-		return rc;
+	cmp = strncmp(a->subtype_name, b->subtype_name, sizeof(a->subtype_name));
+	if (cmp)
+		return cmp;
+	cmp = OSMO_CMP(a->rate, b->rate);
+	if (cmp)
+		return cmp;
+	if (cmp_fmtp) {
+		cmp = strncmp(a->fmtp, b->fmtp, sizeof(a->fmtp));
+		if (cmp)
+			return cmp;
+	}
+	if (cmp_payload_type) {
+		cmp = OSMO_CMP(a->payload_type, b->payload_type);
+		if (cmp)
+			return cmp;
+	}
+	return 0;
+}
 
-	if (a->rate < b->rate)
+/* Compare two lists of audio codecs, returning typical cmp result: 0 on match, and -1 / 1 on mismatch.
+ * The ordering in the two lists may differ, except that the first codec in 'a' must also be the first codec in 'b'.
+ * This is because the first codec typically expresses the preferred codec to use.
+ * If cmp_fmtp is false, do *not* compare the fmtp strings; if true, compare fmtp 1:1 as strings.
+ * If cmp_payload_type is false, do *not* compare the payload_type numbers.
+ * The fmtp is only string-compared -- e.g. if AMR parameters appear in a different order, it amounts to a mismatch even
+ * though all parameters are the same. */
+int sdp_audio_codecs_cmp(const struct sdp_audio_codecs *a, const struct sdp_audio_codecs *b,
+			 bool cmp_fmtp, bool cmp_payload_type)
+{
+	const struct sdp_audio_codec *codec_a;
+	const struct sdp_audio_codec *codec_b;
+	int cmp;
+	if (a == b)
+		return 0;
+	if (!a)
 		return -1;
-	if (a->rate > b->rate)
+	if (!b)
 		return 1;
 
-	rc = strncmp(a->fmtp, b->fmtp, sizeof(a->fmtp));
-	if (rc)
-		return rc;
+	cmp = OSMO_CMP(a->count, b->count);
+	if (cmp)
+		return cmp;
+
+	if (!a->count)
+		return 0;
+
+	/* The first codec is the "chosen" codec and should match. The others may appear in different order. */
+	cmp = sdp_audio_codec_cmp(&a->codec[0], &b->codec[0], cmp_fmtp, cmp_payload_type);
+	if (cmp)
+		return cmp;
+
+	/* See if each codec in a is also present in b */
+	foreach_sdp_audio_codec(codec_a, a) {
+		bool match_found = false;
+		foreach_sdp_audio_codec(codec_b, b) {
+			if (!sdp_audio_codec_cmp(codec_a, codec_b, cmp_fmtp, cmp_payload_type)) {
+				match_found = true;
+				break;
+			}
+		}
+		if (!match_found)
+			return -1;
+	}
 
 	return 0;
 }
@@ -63,13 +121,13 @@ int sdp_audio_codec_cmp(const struct sdp_audio_codec *a, const struct sdp_audio_
  * The payload_type must exist in sdp_msg_payload_type_names.
  * Return the audio codec created or already existing for this payload type number.
  */
-struct sdp_audio_codec *sdp_audio_codec_add(struct sdp_audio_codecs *ac, unsigned int payload_type,
-					    const char *subtype_name, unsigned int rate, const char *fmtp)
+struct sdp_audio_codec *sdp_audio_codecs_add(struct sdp_audio_codecs *ac, unsigned int payload_type,
+					     const char *subtype_name, unsigned int rate, const char *fmtp)
 {
 	struct sdp_audio_codec *codec;
 
 	/* Does an entry already exist? */
-	codec = sdp_audio_codec_by_payload_type(ac, payload_type, false);
+	codec = sdp_audio_codecs_by_payload_type(ac, payload_type, false);
 	if (codec) {
 		/* Already exists, sanity check */
 		if (!codec->subtype_name[0])
@@ -88,7 +146,7 @@ struct sdp_audio_codec *sdp_audio_codec_add(struct sdp_audio_codecs *ac, unsigne
 	}
 
 	/* None exists, create codec entry for this payload type number */
-	codec = sdp_audio_codec_by_payload_type(ac, payload_type, true);
+	codec = sdp_audio_codecs_by_payload_type(ac, payload_type, true);
 	/* NULL means unable to add an entry */
 	if (!codec)
 		return NULL;
@@ -99,14 +157,21 @@ struct sdp_audio_codec *sdp_audio_codec_add(struct sdp_audio_codecs *ac, unsigne
 	return codec;
 }
 
-struct sdp_audio_codec *sdp_audio_codec_add_copy(struct sdp_audio_codecs *ac, const struct sdp_audio_codec *codec)
+struct sdp_audio_codec *sdp_audio_codecs_add_copy(struct sdp_audio_codecs *ac, const struct sdp_audio_codec *codec)
 {
-	return sdp_audio_codec_add(ac, codec->payload_type, codec->subtype_name, codec->rate,
-				   codec->fmtp[0] ? codec->fmtp : NULL);
+	return sdp_audio_codecs_add(ac, codec->payload_type, codec->subtype_name, codec->rate,
+				    codec->fmtp[0] ? codec->fmtp : NULL);
 }
 
-struct sdp_audio_codec *sdp_audio_codec_by_payload_type(struct sdp_audio_codecs *ac, unsigned int payload_type,
-							bool create)
+/* Find or create an entry for the given payload_type number in the given list of codecs.
+ * If the given payload_type number is already present in ac, return the first matching entry.
+ * If no such payload_type number is present: a) return NULL if create == false;
+ * b) If create == true, add a mostly empty codec entry to the end of ac with the given payload_type number, and return
+ * the created entry.
+ * If create == true, a NULL return value means that there was no unused entry left in ac to add this payload_type.
+ */
+struct sdp_audio_codec *sdp_audio_codecs_by_payload_type(struct sdp_audio_codecs *ac, unsigned int payload_type,
+							 bool create)
 {
 	struct sdp_audio_codec *codec;
 	foreach_sdp_audio_codec(codec, ac) {
@@ -130,22 +195,22 @@ struct sdp_audio_codec *sdp_audio_codec_by_payload_type(struct sdp_audio_codecs
 	return codec;
 }
 
-/* Return a given sdp_msg's codec entry that matches the subtype_name, rate and fmtp of the given codec, or NULL if no
- * match is found. Comparison is made by sdp_audio_codec_cmp(). */
-struct sdp_audio_codec *sdp_audio_codec_by_descr(struct sdp_audio_codecs *ac, const struct sdp_audio_codec *codec)
+/* Return a given sdp_msg's codec entry that matches the subtype_name and rate of the given codec, or NULL if no
+ * match is found. Comparison is made by sdp_audio_codec_cmp(cmp_payload_type=false). */
+struct sdp_audio_codec *sdp_audio_codecs_by_descr(struct sdp_audio_codecs *ac, const struct sdp_audio_codec *codec)
 {
 	struct sdp_audio_codec *i;
 	foreach_sdp_audio_codec(i, ac) {
-		if (!sdp_audio_codec_cmp(i, codec))
+		if (!sdp_audio_codec_cmp(i, codec, false, false))
 			return i;
 	}
 	return NULL;
 }
 
 /* Remove the codec entry pointed at by 'codec'. 'codec' must point at an entry of 'sdp' (to use an external codec
- * instance, use sdp_audio_codec_by_descr()).
+ * instance, use sdp_audio_codecs_by_descr()).
  * Return 0 on success, -ENOENT if codec does not point at the sdp->codec array. */
-int sdp_audio_codec_remove(struct sdp_audio_codecs *ac, const struct sdp_audio_codec *codec)
+int sdp_audio_codecs_remove(struct sdp_audio_codecs *ac, const struct sdp_audio_codec *codec)
 {
 	struct sdp_audio_codec *i;
 	if ((codec < ac->codec)
@@ -162,13 +227,29 @@ int sdp_audio_codec_remove(struct sdp_audio_codecs *ac, const struct sdp_audio_c
 	return 0;
 }
 
+static const char * const sdp_mode_str[] = {
+	[SDP_MODE_UNSET] = "-",
+	[SDP_MODE_SENDONLY] = "sendonly",
+	[SDP_MODE_RECVONLY] = "recvonly",
+	[SDP_MODE_SENDRECV] = "sendrecv",
+	[SDP_MODE_INACTIVE] = "inactive",
+};
+
 /* Convert struct sdp_msg to the actual SDP protocol representation */
-int sdp_msg_to_str(char *dst, size_t dst_size, const struct sdp_msg *sdp)
+int sdp_msg_to_sdp_str_buf(char *dst, size_t dst_size, const struct sdp_msg *sdp)
 {
 	const struct sdp_audio_codec *codec;
 	struct osmo_strbuf sb = { .buf = dst, .len = dst_size };
-	const char *ip = sdp->rtp.ip[0] ? sdp->rtp.ip : "0.0.0.0";
-	char ipv = osmo_ip_str_type(ip) == AF_INET6 ? '6' : '4';
+	const char *ip;
+	char ipv;
+
+	if (!sdp) {
+		OSMO_STRBUF_PRINTF(sb, "%s", "");
+		return sb.chars_needed;
+	}
+
+	ip = sdp->rtp.ip[0] ? sdp->rtp.ip : "0.0.0.0";
+	ipv = (osmo_ip_str_type(ip) == AF_INET6) ? '6' : '4';
 
 	OSMO_STRBUF_PRINTF(sb,
 			   "v=0\r\n"
@@ -187,17 +268,19 @@ int sdp_msg_to_str(char *dst, size_t dst_size, const struct sdp_msg *sdp)
 
 	/* Add details for all codecs */
 	foreach_sdp_audio_codec(codec, &sdp->audio_codecs) {
-		if (codec->subtype_name[0]) {
-			OSMO_STRBUF_PRINTF(sb, "a=rtpmap:%d %s/%d\r\n", codec->payload_type, codec->subtype_name,
-					   codec->rate > 0? codec->rate : 8000);
-		}
-
+		if (!sdp_audio_codec_is_set(codec))
+			continue;
+		OSMO_STRBUF_PRINTF(sb, "a=rtpmap:%d %s/%d\r\n", codec->payload_type, codec->subtype_name,
+				   codec->rate > 0 ? codec->rate : 8000);
 		if (codec->fmtp[0])
 			OSMO_STRBUF_PRINTF(sb, "a=fmtp:%d %s\r\n", codec->payload_type, codec->fmtp);
 	}
 
 	OSMO_STRBUF_PRINTF(sb, "a=ptime:%d\r\n", sdp->ptime > 0? sdp->ptime : 20);
 
+	if (sdp->mode != SDP_MODE_UNSET && sdp->mode < ARRAY_SIZE(sdp_mode_str))
+		OSMO_STRBUF_PRINTF(sb, "a=%s\r\n", sdp_mode_str[sdp->mode]);
+
 	return sb.chars_needed;
 }
 
@@ -223,11 +306,9 @@ int sdp_parse_attrib(struct sdp_msg *sdp, const char *src)
 #define A_FMTP "fmtp:"
 #define A_PTIME "ptime:"
 #define A_RTCP "rtcp:"
-#define A_SENDRECV "sendrecv"
-#define A_SENDONLY "sendonly"
-#define A_RECVONLY "recvonly"
 
 	if (osmo_str_startswith(src, A_RTPMAP)) {
+		/* "a=rtpmap:3 GSM/8000" */
 		char *audio_name;
 		unsigned int channels = 1;
 		if (sscanf(src, A_RTPMAP "%u", &payload_type) != 1)
@@ -237,7 +318,7 @@ int sdp_parse_attrib(struct sdp_msg *sdp, const char *src)
 		if (!audio_name || audio_name >= sdp_msg_line_end(src))
 			return -EINVAL;
 
-		codec = sdp_audio_codec_by_payload_type(&sdp->audio_codecs, payload_type, true);
+		codec = sdp_audio_codecs_by_payload_type(&sdp->audio_codecs, payload_type, true);
 		if (!codec)
 			return -ENOSPC;
 
@@ -249,6 +330,7 @@ int sdp_parse_attrib(struct sdp_msg *sdp, const char *src)
 	}
 
 	else if (osmo_str_startswith(src, A_FMTP)) {
+		/* "a=fmtp:112 octet-align=1;mode-set=0,1,2,3" */
 		char *fmtp_str;
 		const char *line_end = sdp_msg_line_end(src);
 		if (sscanf(src, A_FMTP "%u", &payload_type) != 1)
@@ -261,7 +343,7 @@ int sdp_parse_attrib(struct sdp_msg *sdp, const char *src)
 		if (fmtp_str >= line_end)
 			return -EINVAL;
 
-		codec = sdp_audio_codec_by_payload_type(&sdp->audio_codecs, payload_type, true);
+		codec = sdp_audio_codecs_by_payload_type(&sdp->audio_codecs, payload_type, true);
 		if (!codec)
 			return -ENOSPC;
 
@@ -270,6 +352,7 @@ int sdp_parse_attrib(struct sdp_msg *sdp, const char *src)
 	}
 
 	else if (osmo_str_startswith(src, A_PTIME)) {
+		/* "a=ptime:20" */
 		if (sscanf(src, A_PTIME "%u", &sdp->ptime) != 1)
 			return -EINVAL;
 
@@ -279,16 +362,24 @@ int sdp_parse_attrib(struct sdp_msg *sdp, const char *src)
 		/* TODO? */
 	}
 
-	else if (osmo_str_startswith(src, A_SENDRECV)) {
-		/* TODO? */
+	else if (osmo_str_startswith(src, sdp_mode_str[SDP_MODE_SENDRECV])) {
+		/* "a=sendrecv" */
+		sdp->mode = SDP_MODE_SENDRECV;
 	}
 
-	else if (osmo_str_startswith(src, A_SENDONLY)) {
-		/* TODO? */
+	else if (osmo_str_startswith(src, sdp_mode_str[SDP_MODE_SENDONLY])) {
+		/* "a=sendonly" */
+		sdp->mode = SDP_MODE_SENDONLY;
 	}
 
-	else if (osmo_str_startswith(src, A_RECVONLY)) {
-		/* TODO? */
+	else if (osmo_str_startswith(src, sdp_mode_str[SDP_MODE_RECVONLY])) {
+		/* "a=recvonly" */
+		sdp->mode = SDP_MODE_RECVONLY;
+	}
+
+	else if (osmo_str_startswith(src, sdp_mode_str[SDP_MODE_INACTIVE])) {
+		/* "a=inactive" */
+		sdp->mode = SDP_MODE_INACTIVE;
 	}
 
 	return 0;
@@ -347,7 +438,7 @@ static int sdp_parse_media_description(struct sdp_msg *sdp, const char *src)
 		if (sscanf(payload_type_str, "%u", &payload_type) < 1)
 			return -EINVAL;
 
-		codec = sdp_audio_codec_by_payload_type(&sdp->audio_codecs, payload_type, true);
+		codec = sdp_audio_codecs_by_payload_type(&sdp->audio_codecs, payload_type, true);
 		if (!codec)
 			return -ENOSPC;
 
@@ -383,7 +474,7 @@ static int sdp_parse_connection_info(struct sdp_msg *sdp, const char *src)
 }
 
 /* Parse SDP string into struct sdp_msg. Return 0 on success, negative on error. */
-int sdp_msg_from_str(struct sdp_msg *sdp, const char *src)
+int sdp_msg_from_sdp_str(struct sdp_msg *sdp, const char *src)
 {
 	const char *pos;
 	*sdp = (struct sdp_msg){};
@@ -452,8 +543,8 @@ next_line:
 }
 
 /* Leave only those codecs in 'ac_dest' that are also present in 'ac_other'.
- * The matching is made by sdp_audio_codec_cmp(), i.e. payload_type numbers are not compared and fmtp parameters are
- * compared 1:1 as plain strings.
+ * The matching is made by sdp_audio_codec_cmp(cmp_payload_type=false), i.e. payload_type numbers are not compared and
+ * fmtp parameters are compared 1:1 as plain strings.
  * If translate_payload_type_numbers has an effect if ac_dest and ac_other have mismatching payload_type numbers for the
  * same SDP codec descriptions. If translate_payload_type_numbers is true, take the payload_type numbers from ac_other.
  * If false, keep payload_type numbers in ac_dest unchanged. */
@@ -466,10 +557,10 @@ void sdp_audio_codecs_intersection(struct sdp_audio_codecs *ac_dest, const struc
 		struct sdp_audio_codec *other;
 		OSMO_ASSERT(i < ARRAY_SIZE(ac_dest->codec));
 
-		other = sdp_audio_codec_by_descr((struct sdp_audio_codecs*)ac_other, codec);
+		other = sdp_audio_codecs_by_descr((struct sdp_audio_codecs *)ac_other, codec);
 
 		if (!other) {
-			OSMO_ASSERT(sdp_audio_codec_remove(ac_dest, codec) == 0);
+			OSMO_ASSERT(sdp_audio_codecs_remove(ac_dest, codec) == 0);
 			i--;
 			continue;
 		}
@@ -505,27 +596,30 @@ void sdp_audio_codecs_select(struct sdp_audio_codecs *ac, struct sdp_audio_codec
 
 /* Short single-line representation of an SDP audio codec, convenient for logging.
  * Like "AMR/8000:octet-align=1#122" */
-int sdp_audio_codec_name_buf(char *buf, size_t buflen, const struct sdp_audio_codec *codec)
+int sdp_audio_codec_to_str_buf(char *buf, size_t buflen, const struct sdp_audio_codec *codec)
 {
 	struct osmo_strbuf sb = { .buf = buf, .len = buflen };
 	OSMO_STRBUF_PRINTF(sb, "%s", codec->subtype_name);
+	if (codec->rate != 8000)
+		OSMO_STRBUF_PRINTF(sb, "/%u", codec->rate);
 	if (codec->fmtp[0])
 		OSMO_STRBUF_PRINTF(sb, ":%s", codec->fmtp);
+	OSMO_STRBUF_PRINTF(sb, "#%d", codec->payload_type);
 	return sb.chars_needed;
 }
 
-char *sdp_audio_codec_name_c(void *ctx, const struct sdp_audio_codec *codec)
+char *sdp_audio_codec_to_str_c(void *ctx, const struct sdp_audio_codec *codec)
 {
-	OSMO_NAME_C_IMPL(ctx, 32, "sdp_audio_codec_name_c-ERROR", sdp_audio_codec_name_buf, codec)
+	OSMO_NAME_C_IMPL(ctx, 32, "sdp_audio_codec_to_str_c-ERROR", sdp_audio_codec_to_str_buf, codec)
 }
 
-const char *sdp_audio_codec_name(const struct sdp_audio_codec *codec)
+const char *sdp_audio_codec_to_str(const struct sdp_audio_codec *codec)
 {
-	return sdp_audio_codec_name_c(OTC_SELECT, codec);
+	return sdp_audio_codec_to_str_c(OTC_SELECT, codec);
 }
 
 /* Short single-line representation of a list of SDP audio codecs, convenient for logging */
-int sdp_audio_codecs_name_buf(char *buf, size_t buflen, const struct sdp_audio_codecs *ac)
+int sdp_audio_codecs_to_str_buf(char *buf, size_t buflen, const struct sdp_audio_codecs *ac)
 {
 	struct osmo_strbuf sb = { .buf = buf, .len = buflen };
 	const struct sdp_audio_codec *codec;
@@ -535,23 +629,23 @@ int sdp_audio_codecs_name_buf(char *buf, size_t buflen, const struct sdp_audio_c
 		bool first = (codec == ac->codec);
 		if (!first)
 			OSMO_STRBUF_PRINTF(sb, ",");
-		OSMO_STRBUF_APPEND(sb, sdp_audio_codec_name_buf, codec);
+		OSMO_STRBUF_APPEND(sb, sdp_audio_codec_to_str_buf, codec);
 	}
 	return sb.chars_needed;
 }
 
-char *sdp_audio_codecs_name_c(void *ctx, const struct sdp_audio_codecs *ac)
+char *sdp_audio_codecs_to_str_c(void *ctx, const struct sdp_audio_codecs *ac)
 {
-	OSMO_NAME_C_IMPL(ctx, 128, "sdp_audio_codecs_name_c-ERROR", sdp_audio_codecs_name_buf, ac)
+	OSMO_NAME_C_IMPL(ctx, 128, "sdp_audio_codecs_to_str_c-ERROR", sdp_audio_codecs_to_str_buf, ac)
 }
 
-const char *sdp_audio_codecs_name(const struct sdp_audio_codecs *ac)
+const char *sdp_audio_codecs_to_str(const struct sdp_audio_codecs *ac)
 {
-	return sdp_audio_codecs_name_c(OTC_SELECT, ac);
+	return sdp_audio_codecs_to_str_c(OTC_SELECT, ac);
 }
 
 /* Short single-line representation of an SDP message, convenient for logging */
-int sdp_msg_name_buf(char *buf, size_t buflen, const struct sdp_msg *sdp)
+int sdp_msg_to_str_buf(char *buf, size_t buflen, const struct sdp_msg *sdp)
 {
 	struct osmo_strbuf sb = { .buf = buf, .len = buflen };
 	if (!sdp) {
@@ -561,17 +655,17 @@ int sdp_msg_name_buf(char *buf, size_t buflen, const struct sdp_msg *sdp)
 
 	OSMO_STRBUF_PRINTF(sb, OSMO_SOCKADDR_STR_FMT, OSMO_SOCKADDR_STR_FMT_ARGS(&sdp->rtp));
 	OSMO_STRBUF_PRINTF(sb, "{");
-	OSMO_STRBUF_APPEND(sb, sdp_audio_codecs_name_buf, &sdp->audio_codecs);
+	OSMO_STRBUF_APPEND(sb, sdp_audio_codecs_to_str_buf, &sdp->audio_codecs);
 	OSMO_STRBUF_PRINTF(sb, "}");
 	return sb.chars_needed;
 }
 
-char *sdp_msg_name_c(void *ctx, const struct sdp_msg *sdp)
+char *sdp_msg_to_str_c(void *ctx, const struct sdp_msg *sdp)
 {
-	OSMO_NAME_C_IMPL(ctx, 128, "sdp_msg_name_c-ERROR", sdp_msg_name_buf, sdp)
+	OSMO_NAME_C_IMPL(ctx, 128, "sdp_msg_to_str_c-ERROR", sdp_msg_to_str_buf, sdp)
 }
 
-const char *sdp_msg_name(const struct sdp_msg *sdp)
+const char *sdp_msg_to_str(const struct sdp_msg *sdp)
 {
-	return sdp_msg_name_c(OTC_SELECT, sdp);
+	return sdp_msg_to_str_c(OTC_SELECT, sdp);
 }
