@@ -1007,6 +1007,7 @@ void ms_sends_assignment_complete(enum mgcp_codecs assigned_codec)
 	ran_dec = (struct ran_msg){
 		.msg_type = RAN_MSG_ASSIGNMENT_COMPLETE,
 		.assignment_complete = {
+			.codec_present = true,
 			.codec = assigned_codec,
 		},
 	};
@@ -1135,6 +1136,7 @@ static void run_tests(int nr)
 struct gsm_network *test_net(void *ctx)
 {
 	struct gsm_network *net = gsm_network_init(ctx, mncc_recv);
+	struct mgcp_client *client;
 
 	net->gsup_server_addr_str = talloc_strdup(net, "no_gsup_server");
 	net->gsup_server_port = 0;
@@ -1171,8 +1173,9 @@ struct gsm_network *test_net(void *ctx)
 	net->mgw.tdefs = g_mgw_tdefs;
 	mgcp_client_conf_init(&net->mgw.conf);
 	net->mgw.tdefs = g_mgw_tdefs;
-	net->mgw.client = mgcp_client_init(net, &net->mgw.conf);
-
+	net->mgw.mgw_pool = mgcp_client_pool_alloc(net);
+	client = mgcp_client_init(net, &net->mgw.conf);
+	mgcp_client_pool_register_single(net->mgw.mgw_pool, client);
 	return net;
 }
 
